@@ -1,12 +1,31 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import re
 from binascii import hexlify, unhexlify
+from collections.abc import Generator
+from typing import Optional, cast
+
+from kitty.options.types import Options
 
 
-names = 'xterm-kitty', 'KovIdTTY'
+def modify_key_bytes(keybytes: bytes, amt: int) -> bytes:
+    if amt == 0:
+        return keybytes
+    ans = bytearray(keybytes)
+    samt = str(amt).encode('ascii')
+    if ans[-1] == ord('~'):
+        return bytes(ans[:-1] + bytearray(b';' + samt + b'~'))
+    if ans[1] == ord('O'):
+        return bytes(ans[:1] + bytearray(b'[1;' + samt) + ans[-1:])
+    raise ValueError(f'Unknown key type in key: {keybytes!r}')
+
+
+def encode_keystring(keybytes: bytes) -> str:
+    return keybytes.decode('ascii').replace('\033', r'\E')
+
+
+names = Options.term, 'KovIdTTY'
 
 termcap_aliases = {
     'TN': 'name'
@@ -34,13 +53,14 @@ bool_capabilities = {
     # Terminfo extension used by tmux to detect true color support (non-standard)
     'Tc',
     # Indicates support for styled and colored underlines (non-standard) as
-    # described at:
-    # https://github.com/kovidgoyal/kitty/blob/master/protocol-extensions.asciidoc
+    # described at: https://sw.kovidgoyal.net/kitty/underlines/
     'Su',
     # Indicates support for full keyboard mode (non-standard) as
     # described at:
     # https://github.com/kovidgoyal/kitty/blob/master/protocol-extensions.asciidoc
     'fullkbd',
+    # Terminal supports focus events: https://lists.gnu.org/archive/html/bug-ncurses/2023-10/msg00117.html
+    'XF',
 
     # The following are entries that we don't use
     # # background color erase
@@ -92,8 +112,12 @@ string_capabilities = {
     'civis': r'\E[?25l',
     # Clear screen
     'clear': r'\E[H\E[2J',
+    # Clear scrollback. This is disabled because the clear program on Linux by default, not as
+    # an option, uses it and nukes the scrollback. What's more this behavior was silently changed
+    # around 2013. Given clear is maintained as part of ncurses this kind of crap is no surprise.
+    # 'E3': r'\E[3J',
     # Make cursor appear normal
-    'cnorm': r'\E[?12l\E[?25h',
+    'cnorm': r'\E[?12h\E[?25h',
     # Carriage return
     'cr': r'^M',  # CR (carriage return \r)
     # Change scroll region
@@ -156,60 +180,23 @@ string_capabilities = {
     # 'invis': r'\E[8m',
     # Backspace
     'kbs': r'\177',
-    # Left
-    'kcub1': r'\EOD',
-    # Down
-    'kcud1': r'\EOB',
-    # Right
-    'kcuf1': r'\EOC',
-    # Up
-    'kcuu1': r'\EOA',
-    # Function keys
-    'kf1': r'\EOP',
-    'kf2': r'\EOQ',
-    'kf3': r'\EOR',
-    'kf4': r'\EOS',
-    'kf5': r'\E[15~',
-    'kf6': r'\E[17~',
-    'kf7': r'\E[18~',
-    'kf8': r'\E[19~',
-    'kf9': r'\E[20~',
-    'kf10': r'\E[21~',
-    'kf11': r'\E[23~',
-    'kf12': r'\E[24~',
-    'kf13': r'\E[1;2P',
-    'kf14': r'\E[1;2Q',
-    'kf15': r'\E[1;2R',
-    'kf16': r'\E[1;2S',
-    'kf17': r'\E[15;2~',
-    'kf18': r'\E[17;2~',
-    'kf19': r'\E[18;2~',
-    'kf20': r'\E[19;2~',
-    'kf21': r'\E[20;2~',
-    'kf22': r'\E[21;2~',
-    'kf23': r'\E[23;2~',
-    'kf24': r'\E[24;2~',
-    'kf25': r'\E[1;5P',
-    # Home
-    'khome': r'\EOH',
-    # End
-    'kend': r'\EOF',
-    # Insert character key
-    'kich1': r'\E[2~',
-    # Delete character key
-    'kdch1': r'\E[3~',
     # Mouse event has occurred
     'kmous': r'\E[M',
-    # Page down
-    'knp': r'\E[6~',
-    # Page up
-    'kpp': r'\E[5~',
+
+    # These break mouse events in htop so they are disabled
+    # Turn on mouse reporting
+    # 'XM': '\E[?1006;1004;1000%?%p1%{1}%=%th%el%;',
+    # Expected format for mouse reporting escape codes
+    # 'xm': r'\E[<%i%p3%d;%p1%d;%p2%d;%?%p4%tM%em%;',
     # Scroll backwards (reverse index)
+
     'kri': r'\E[1;2A',
     # scroll forwards (index)
     'kind': r'\E[1;2B',
     # Restore cursor
     'rc': r'\E8',
+    # Repeat preceding character
+    'rep': r'%p1%c\E[%p2%{1}%-%db',
     # Reverse video
     'rev': r'\E[7m',
     # Scroll backwards the specified number of lines (reverse index)
@@ -227,6 +214,8 @@ string_capabilities = {
     'rmso': r'\E[27m',
     # Exit underline mode
     'rmul': r'\E[24m',
+    # Exit strikethrough mode
+    'rmxx': r'\E[29m',
     # Reset string1 (empty OSC sequence to exit OSC/OTH modes, and regular reset)
     'rs1': r'\E]\E\\\Ec',
     # Save cursor
@@ -245,7 +234,7 @@ string_capabilities = {
     'smam': r'\E[?7h',
     # Start alternate screen
     'smcup': r'\E[?1049h',
-    # Enster insert mode
+    # Enter insert mode
     'smir': r'\E[4h',
     # Enter application keymap mode
     'smkx': r'\E[?1h',
@@ -253,6 +242,10 @@ string_capabilities = {
     'smso': r'\E[7m',
     # Enter underline mode
     'smul': r'\E[4m',
+    'Smulx': r'\E[4:%p1%dm',  # this is a non-standard extension that some terminals use, so match them
+    # Enter strikethrough mode
+    'smxx': r'\E[9m',
+    'Sync': r'\EP=%p1%ds\E\\',  # this is a non-standard extension supported by tmux for synchronized updates
     # Clear all tab stops
     'tbc': r'\E[3g',
     # To status line (used to set window titles)
@@ -260,7 +253,7 @@ string_capabilities = {
     # From status line (end window title string)
     'fsl': r'^G',
     # Disable status line (clear window title)
-    'dsl': r'\E]2;\007',
+    'dsl': r'\E]2;\E\\',
     # Move to specified line
     'vpa': r'\E[%i%p1%dd',
     # Enter italics mode
@@ -270,11 +263,6 @@ string_capabilities = {
     # Select alternate charset
     'smacs': r'\E(0',
     'rmacs': r'\E(B',
-    # Shifted keys
-    'kRIT': r'\E[1;2C',
-    'kLFT': r'\E[1;2D',
-    'kEND': r'\E[1;2F',
-    'kHOM': r'\E[1;2H',
     # Special keys
     'khlp': r'',
     'kund': r'',
@@ -286,6 +274,48 @@ string_capabilities = {
     'setrgbf': r'\E[38:2:%p1%d:%p2%d:%p3%dm',
     # Set RGB background color (non-standard used by neovim)
     'setrgbb': r'\E[48:2:%p1%d:%p2%d:%p3%dm',
+    # DECSCUSR Set cursor style
+    'Ss': r'\E[%p1%d\sq',
+    # DECSCUSR Reset cursor style to power-on default
+    'Se': r'\E[2\sq',
+    # Set cursor color
+    'Cs': r'\E]12;%p1%s\007',
+    # Reset cursor color
+    'Cr': r'\E]112\007',
+    # Indicates support for styled and colored underlines (non-standard) as
+    # described at: https://sw.kovidgoyal.net/kitty/underlines/
+    # 'Setulc' is equivalent to the 'Su' boolean capability. Until
+    # standardized, specify both for application compatibility.
+    'Setulc': r'\E[58:2:%p1%{65536}%/%d:%p1%{256}%/%{255}%&%d:%p1%{255}%&%d%;m',
+
+    # The following entries are for compatibility with xterm,
+    # and shell scripts using e.g. `tput u7` to emit a CPR escape
+    # See https://invisible-island.net/ncurses/terminfo.src.html
+    # and INTERPRETATION OF USER CAPABILITIES
+    'u6': r'\E[%i%d;%dR',
+    'u7': r'\E[6n',
+    'u8': r'\E[?%[;0123456789]c',
+    'u9': r'\E[c',
+
+    # Bracketed paste, added to ncurses 6.4 in 2023
+    'PS': r'\E[200~',
+    'PE': r'\E[201~',
+    'BE': r'\E[?2004h',
+    'BD': r'\E[?2004l',
+
+    # XTVERSION
+    'XR': r'\E[>0q',
+    # OSC 52 clipboard access
+    'Ms': r'\E]52;%p1%s;%p2%s\E\\',
+    # Send device attributes (report version)
+    'RV': r'\E[>c',
+    # Focus In and Out events
+    'kxIN': r'\E[I',
+    'kxOUT': r'\E[O',
+    # Enable/disable focus reporting
+    # Add to ncurses in: https://lists.gnu.org/archive/html/bug-ncurses/2023-10/msg00117.html
+    'fe': r'\E[?1004h',
+    'fd': r'\E[?1004l',
 
     # The following are entries that we don't use
     # # turn on blank mode, (characters invisible)
@@ -298,6 +328,34 @@ string_capabilities = {
     # 'rs2': r'\E[!p\E[?3;4l\E[4l\E>',
 }
 
+string_capabilities.update({
+    f'kf{n}':
+        encode_keystring(modify_key_bytes(b'\033' + value, 0))
+    for n, value in zip(range(1, 13),
+                        b'OP OQ OR OS [15~ [17~ [18~ [19~ [20~ [21~ [23~ [24~'.split())
+})
+
+string_capabilities.update({
+    f'kf{offset + n}':
+        encode_keystring(modify_key_bytes(b'\033' + value, mod))
+    for offset, mod in {12: 2, 24: 5, 36: 6, 48: 3, 60: 4}.items()
+    for n, value in zip(range(1, 13),
+                        b'OP OQ [13~ OS [15~ [17~ [18~ [19~ [20~ [21~ [23~ [24~'.split())
+    if offset + n < 64
+})
+
+string_capabilities.update({
+    name.format(unmod=unmod, key=key):
+        encode_keystring(modify_key_bytes(b'\033' + value, mod))
+    for unmod, key, value in zip(
+        'cuu1 cud1 cuf1 cub1 beg end home ich1 dch1 pp  np'.split(),
+        'UP   DN   RIT  LFT  BEG END HOM  IC   DC   PRV NXT'.split(),
+        b'OA  OB   OC   OD   OE  OF  OH   [2~  [3~  [5~ [6~'.split())
+    for name, mod in {
+        'k{unmod}': 0, 'k{key}': 2, 'k{key}3': 3, 'k{key}4': 4,
+        'k{key}5': 5, 'k{key}6': 6, 'k{key}7': 7}.items()
+})
+
 termcap_aliases.update({
     'ac': 'acsc',
     'bl': 'bel',
@@ -346,31 +404,6 @@ termcap_aliases.update({
     'kd': 'kcud1',
     'kr': 'kcuf1',
     'ku': 'kcuu1',
-    'k1': 'kf1',
-    'k2': 'kf2',
-    'k3': 'kf3',
-    'k4': 'kf4',
-    'k5': 'kf5',
-    'k6': 'kf6',
-    'k7': 'kf7',
-    'k8': 'kf8',
-    'k9': 'kf9',
-    'k;': 'kf10',
-    'F1': 'kf11',
-    'F2': 'kf12',
-    'F3': 'kf13',
-    'F4': 'kf14',
-    'F5': 'kf15',
-    'F6': 'kf16',
-    'F7': 'kf17',
-    'F8': 'kf18',
-    'F9': 'kf19',
-    'FA': 'kf20',
-    'FB': 'kf21',
-    'FC': 'kf22',
-    'FD': 'kf23',
-    'FE': 'kf24',
-    'FF': 'kf25',
     'kh': 'khome',
     '@7': 'kend',
     'kI': 'kich1',
@@ -381,6 +414,7 @@ termcap_aliases.update({
     'kR': 'kri',
     'kF': 'kind',
     'rc': 'rc',
+    'rp': 'rep',
     'mr': 'rev',
     'sr': 'ri',
     'SR': 'rin',
@@ -389,6 +423,7 @@ termcap_aliases.update({
     'ei': 'rmir',
     'se': 'rmso',
     'ue': 'rmul',
+    'Te': 'rmxx',
     'r1': 'rs1',
     'sc': 'sc',
     'AB': 'setab',
@@ -401,6 +436,7 @@ termcap_aliases.update({
     'im': 'smir',
     'so': 'smso',
     'us': 'smul',
+    'Ts': 'smxx',
     'ct': 'tbc',
     'cv': 'vpa',
     'ZH': 'sitm',
@@ -410,8 +446,12 @@ termcap_aliases.update({
     'ks': 'smkx',
     'ke': 'rmkx',
     '#2': 'kHOM',
+    '#3': 'kIC',
     '#4': 'kLFT',
+    '*4': 'kDC',
     '*7': 'kEND',
+    '%c': 'kNXT',
+    '%e': 'kPRV',
     '%i': 'kRIT',
     '%1': 'khlp',
     '&8': 'kund',
@@ -423,6 +463,10 @@ termcap_aliases.update({
     'fs': 'fsl',
     'ds': 'dsl',
 
+    'u6': 'u6',
+    'u7': 'u7',
+    'u8': 'u8',
+    'u9': 'u9',
 
     # 'ut': 'bce',
     # 'ds': 'dsl',
@@ -433,21 +477,33 @@ termcap_aliases.update({
     # 'r2': 'rs2',
 })
 
-queryable_capabilities = numeric_capabilities.copy()
+termcap_aliases.update({
+    tc: f'kf{n}'
+    for n, tc in enumerate(
+        'k1 k2 k3 k4 k5 k6 k7 k8 k9 k; F1 F2 F3 F4 F5 F6 F7 F8 F9 FA '
+        'FB FC FD FE FF FG FH FI FJ FK FL FM FN FO FP FQ FR FS FT FU '
+        'FV FW FX FY FZ Fa Fb Fc Fd Fe Ff Fg Fh Fi Fj Fk Fl Fm Fn Fo '
+        'Fp Fq Fr'.split(), 1)})
+
+queryable_capabilities = cast(dict[str, str], numeric_capabilities.copy())
 queryable_capabilities.update(string_capabilities)
 extra = (bool_capabilities | numeric_capabilities.keys() | string_capabilities.keys()) - set(termcap_aliases.values())
-no_termcap_for = frozenset('Su Tc setrgbf setrgbb fullkbd'.split())
+no_termcap_for = frozenset(
+    'XR XM xm Ms RV kxIN kxOUT Cr Cs Se Ss Setulc Su Smulx Sync Tc PS PE BE BD setrgbf setrgbb fullkbd kUP kDN kbeg kBEG fe fd XF'.split() + [
+        f'k{key}{mod}'
+        for key in 'UP DN RIT LFT BEG END HOM IC DC PRV NXT'.split()
+        for mod in range(3, 8)])
 if extra - no_termcap_for:
-    raise Exception('Termcap aliases not complete, missing: {}'.format(extra - no_termcap_for))
+    raise Exception(f'Termcap aliases not complete, missing: {extra - no_termcap_for}')
 del extra
 
 
-def generate_terminfo():
+def generate_terminfo() -> str:
     # Use ./build-terminfo to update definition files
     ans = ['|'.join(names)]
     ans.extend(sorted(bool_capabilities))
-    ans.extend('{}#{}'.format(k, numeric_capabilities[k]) for k in sorted(numeric_capabilities))
-    ans.extend('{}={}'.format(k, string_capabilities[k]) for k in sorted(string_capabilities))
+    ans.extend(f'{k}#{numeric_capabilities[k]}' for k in sorted(numeric_capabilities))
+    ans.extend(f'{k}={string_capabilities[k]}' for k in sorted(string_capabilities))
     return ',\n\t'.join(ans) + ',\n'
 
 
@@ -455,35 +511,51 @@ octal_escape = re.compile(r'\\([0-7]{3})')
 escape_escape = re.compile(r'\\[eE]')
 
 
-def key_as_bytes(name):
+def key_as_bytes(name: str) -> bytes:
     ans = string_capabilities[name]
     ans = octal_escape.sub(lambda m: chr(int(m.group(1), 8)), ans)
     ans = escape_escape.sub('\033', ans)
     return ans.encode('ascii')
 
 
-def get_capabilities(query_string):
+def get_capabilities(query_string: str, opts: 'Options', window_id: int = 0, os_window_id: int = 0) -> Generator[str, None, None]:
     from .fast_data_types import ERROR_PREFIX
-    ans = []
-    try:
-        for q in query_string.split(';'):
-            name = qname = unhexlify(q).decode('utf-8')
-            if name in ('TN', 'name'):
-                val = names[0]
+
+    def result(encoded_query_name: str, x: Optional[str] = None) -> str:
+        if not x:
+            valid = 0 if x is None else 1
+            return f'{valid}+r{encoded_query_name}'
+        return f'1+r{encoded_query_name}={hexlify(str(x).encode("utf-8")).decode("ascii")}'
+
+    for encoded_query_name in query_string.split(';'):
+        name = qname = unhexlify(encoded_query_name).decode('utf-8')
+        if name in ('TN', 'name'):
+            yield result(encoded_query_name, names[0])
+        elif name.startswith('kitty-query-'):
+            from kittens.query_terminal.main import get_result
+            name = name[len('kitty-query-'):]
+            rval = get_result(name, window_id, os_window_id)
+            if rval is None:
+                from .utils import log_error
+                log_error('Unknown kitty terminfo query:', name)
+                yield result(encoded_query_name)
             else:
+                yield result(encoded_query_name, rval)
+        else:
+            if name in bool_capabilities:
+                yield result(encoded_query_name, '')
+                continue
+            try:
+                val = queryable_capabilities[name]
+            except KeyError:
                 try:
-                    val = queryable_capabilities[name]
-                except KeyError:
-                    try:
-                        qname = termcap_aliases[name]
-                        val = queryable_capabilities[qname]
-                    except Exception:
-                        from .utils import log_error
-                        log_error(ERROR_PREFIX, 'Unknown terminfo property:', name)
-                        raise
-                if qname in string_capabilities and '%' not in val:
-                    val = key_as_bytes(qname).decode('ascii')
-            ans.append(q + '=' + hexlify(str(val).encode('utf-8')).decode('ascii'))
-        return '1+r' + ';'.join(ans)
-    except Exception:
-        return '0+r' + query_string
+                    qname = termcap_aliases[name]
+                    val = queryable_capabilities[qname]
+                except Exception:
+                    from .utils import log_error
+                    log_error(ERROR_PREFIX, 'Unknown terminfo property:', name)
+                    yield result(encoded_query_name)
+                    continue
+            if qname in string_capabilities and '%' not in val:
+                val = key_as_bytes(qname).decode('ascii')
+            yield result(encoded_query_name, val)
